@@ -13,6 +13,8 @@
 #include "Sndfile.h"
 #ifdef MODPLUG_TRACKER
 #include "../mptrack/Mptrack.h"
+#include "mpt/io_file/inputfile.hpp"
+#include "mpt/io_file_read/inputfile_filecursor.hpp"
 #include "../common/mptFileIO.h"
 #endif
 #include "Dlsbank.h"
@@ -158,6 +160,9 @@ enum DLSArt : uint32
 	ART_PITCH_EG_RELEASETIME  = DLSArt(CONN_SRC_NONE, CONN_SRC_NONE, CONN_DST_EG2_RELEASETIME),
 	ART_PITCH_EG_VELTOATTACK  = DLSArt(CONN_SRC_KEYONVELOCITY, CONN_SRC_NONE, CONN_DST_EG2_ATTACKTIME),
 	ART_PITCH_EG_KEYTODECAY   = DLSArt(CONN_SRC_KEYNUMBER,     CONN_SRC_NONE, CONN_DST_EG2_DECAYTIME),
+	ART_PITCH_EG_DELAYTIME    = DLSArt(CONN_SRC_NONE, CONN_SRC_NONE, CONN_DST_EG2_DELAYTIME),
+	ART_PITCH_EG_HOLDTIME     = DLSArt(CONN_SRC_NONE, CONN_SRC_NONE, CONN_DST_EG2_HOLDTIME),
+	ART_PITCH_EG_DEPTH        = DLSArt(CONN_SRC_EG2,  CONN_SRC_NONE, CONN_DST_PITCH),
 
 	// Default Pan
 	ART_DEFAULTPAN = DLSArt(CONN_SRC_NONE, CONN_SRC_NONE, CONN_DST_PAN),
@@ -355,11 +360,18 @@ enum SF2Generators : uint16
 {
 	SF2_GEN_START_LOOP_FINE		= 2,
 	SF2_GEN_END_LOOP_FINE		= 3,
+	SF2_GEN_MODENVTOPITCH		= 7,
 	SF2_GEN_MODENVTOFILTERFC	= 11,
 	SF2_GEN_PAN					= 17,
+	SF2_GEN_DELAYMODENV			= 25,
+	SF2_GEN_ATTACKMODENV		= 26,
+	SF2_GEN_HOLDMODENV			= 27,
 	SF2_GEN_DECAYMODENV			= 28,
+	SF2_GEN_SUSTAINMODENV		= 29,
+	SF2_GEN_RELEASEMODENV		= 30,
+	SF2_GEN_DELAYVOLENV			= 33,
 	SF2_GEN_ATTACKVOLENV		= 34,
-	SF2_GEN_HOLDVOLENV			= 34,
+	SF2_GEN_HOLDVOLENV			= 35,
 	SF2_GEN_DECAYVOLENV			= 36,
 	SF2_GEN_SUSTAINVOLENV		= 37,
 	SF2_GEN_RELEASEVOLENV		= 38,
@@ -405,6 +417,8 @@ struct SFGenList
 {
 	uint16le sfGenOper;
 	uint16le genAmount;
+
+	bool ApplyToEnvelope(DLSENVELOPE &env) const;
 };
 
 MPT_BINARY_STRUCT(SFGenList, 4)
@@ -425,13 +439,7 @@ struct SFInstBag
 
 MPT_BINARY_STRUCT(SFInstBag, 4)
 
-struct SFInstGenList
-{
-	uint16le sfGenOper;
-	uint16le genAmount;
-};
-
-MPT_BINARY_STRUCT(SFInstGenList, 4)
+using SFInstGenList = SFGenList;
 
 struct SFSample
 {
@@ -479,6 +487,13 @@ static uint8 DLSSustainLevelToLinear(int32 sustain)
 }
 
 
+static int16 SF2TimeToDLS(int16 amount)
+{
+	int32 time = CDLSBank::DLS32BitTimeCentsToMilliseconds(static_cast<int32>(amount) << 16);
+	return static_cast<int16>(Clamp(time, 20, 20000) / 20);
+}
+
+
 static uint8 SF2SustainLevelToLinear(int32 sustain)
 {
 	// 0.1% units
@@ -487,6 +502,60 @@ static uint8 SF2SustainLevelToLinear(int32 sustain)
 }
 
 
+bool SFGenList::ApplyToEnvelope(DLSENVELOPE &env) const
+{
+	switch(sfGenOper)
+	{
+	case SF2_GEN_DELAYVOLENV:
+		env.volumeEnv.delay = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_ATTACKVOLENV:
+		env.volumeEnv.attack = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_HOLDVOLENV:
+		env.volumeEnv.hold = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_DECAYVOLENV:
+		env.volumeEnv.decay = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_SUSTAINVOLENV:
+		// 0.1% units
+		if(genAmount >= 0)
+		{
+			env.volumeEnv.sustainLevel = SF2SustainLevelToLinear(genAmount);
+		}
+		break;
+	case SF2_GEN_RELEASEVOLENV:
+		env.volumeEnv.release = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_DELAYMODENV:
+		env.pitchEnv.delay = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_ATTACKMODENV:
+		env.pitchEnv.attack = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_HOLDMODENV:
+		env.pitchEnv.hold = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_DECAYMODENV:
+		env.pitchEnv.decay = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_SUSTAINMODENV:
+		env.pitchEnv.sustainLevel = SF2SustainLevelToLinear(genAmount);
+		break;
+	case SF2_GEN_RELEASEMODENV:
+		env.pitchEnv.release = SF2TimeToDLS(genAmount);
+		break;
+	case SF2_GEN_MODENVTOPITCH:
+		env.pitchEnvDepth = static_cast<int16>(genAmount);
+		break;
+
+	default: return false;
+	}
+	return true;
+}
+
+
 int32 CDLSBank::DLS32BitTimeCentsToMilliseconds(int32 lTimeCents)
 {
 	// tc = log2(time[secs]) * 1200*65536
@@ -499,6 +568,20 @@ int32 CDLSBank::DLS32BitTimeCentsToMilliseconds(int32 lTimeCents)
 }
 
 
+uint16 CDLSBank::DLSEnvelopeTimeCentsToMilliseconds(int32 value)
+{
+	if(value <= -0x40000000)
+		return 0;
+	
+	int32 decaytime = DLS32BitTimeCentsToMilliseconds(value);
+	if(decaytime > 20000)
+		decaytime = 20000;
+	if(decaytime >= 20)
+		return static_cast<uint16>(decaytime / 20);
+	return 0;
+}
+
+
 // 0dB = 0x10000
 int32 CDLSBank::DLS32BitRelativeGainToLinear(int32 lCentibels)
 {
@@ -584,51 +667,65 @@ bool CDLSBank::IsDLSBank(const mpt::PathString &filename)
 
 const DLSINSTRUMENT *CDLSBank::FindInstrument(bool isDrum, uint32 bank, uint32 program, uint32 key, uint32 *pInsNo) const
 {
-	// This helps finding the "more correct" instrument if we search for an instrument in any bank, and the higher-bank instruments appear first in the file
-	// Fixes issues when loading GeneralUser GS into OpenMPT's MIDI library.
-	std::vector<std::reference_wrapper<const DLSINSTRUMENT>> sortedInstr{m_Instruments.begin(), m_Instruments.end()};
-	if(bank >= 0x4000 || program >= 0x80)
+	uint32 minBank = ((bank << 1) & 0x7F00) | (bank & 0x7F);
+	uint32 maxBank = minBank;
+	if(bank >= 0x4000)
 	{
-		std::sort(sortedInstr.begin(), sortedInstr.end(), [](const DLSINSTRUMENT &l, const DLSINSTRUMENT &r)
-				{ return std::tie(l.ulBank, l.ulInstrument) < std::tie(r.ulBank, r.ulInstrument); });
+		minBank = 0x0000;
+		maxBank = 0x7F7F;
+	}
+	if(isDrum)
+	{
+		minBank |= F_INSTRUMENT_DRUMS;
+		maxBank |= F_INSTRUMENT_DRUMS;
 	}
 
-	for(const DLSINSTRUMENT &dlsIns : sortedInstr)
+	const bool singleInstr = (minBank == maxBank) && (program < 0x80);
+	const auto CompareInstrFunc = [singleInstr](const DLSINSTRUMENT &l, const DLSINSTRUMENT &r)
 	{
-		uint32 insbank = ((dlsIns.ulBank & 0x7F00) >> 1) | (dlsIns.ulBank & 0x7F);
-		if((bank >= 0x4000) || (insbank == bank))
-		{
-			if(isDrum && (dlsIns.ulBank & F_INSTRUMENT_DRUMS))
-			{
-				if((program >= 0x80) || (program == (dlsIns.ulInstrument & 0x7F)))
-				{
-					for(const auto &region : dlsIns.Regions)
-					{
-						if(region.IsDummy())
-							continue;
+		if(singleInstr)
+			return l < r;
+		else
+			return l.ulBank < r.ulBank;
+	};
 
-						if((!key || key >= 0x80)
-						   || (key >= region.uKeyMin && key <= region.uKeyMax))
-						{
-							if(pInsNo)
-								*pInsNo = static_cast<uint32>(std::distance(m_Instruments.data(), &dlsIns));
-							// cppcheck false-positive
-							// cppcheck-suppress returnDanglingLifetime
-							return &dlsIns;
-						}
-					}
-				}
-			} else if(!isDrum && !(dlsIns.ulBank & F_INSTRUMENT_DRUMS))
+	DLSINSTRUMENT findInstr{};
+	findInstr.ulInstrument = program;
+	findInstr.ulBank = minBank;
+	const auto minInstr = std::lower_bound(m_Instruments.begin(), m_Instruments.end(), findInstr, CompareInstrFunc);
+	findInstr.ulBank = maxBank;
+	const auto maxInstr = std::upper_bound(m_Instruments.begin(), m_Instruments.end(), findInstr, CompareInstrFunc);
+	const auto instrRange = mpt::as_span(m_Instruments.data() + std::distance(m_Instruments.begin(), minInstr), std::distance(minInstr, maxInstr));
+
+	for(const DLSINSTRUMENT &dlsIns : instrRange)
+	{
+		if((program < 0x80) && program != (dlsIns.ulInstrument & 0x7F))
+			continue;
+
+		if(isDrum)
+		{
+			const bool anyKey = !key || key >= 0x80;
+			for(const auto &region : dlsIns.Regions)
 			{
-				if((program >= 0x80) || (program == (dlsIns.ulInstrument & 0x7F)))
+				if(region.IsDummy())
+					continue;
+
+				if(anyKey || (key >= region.uKeyMin && key <= region.uKeyMax))
 				{
 					if(pInsNo)
 						*pInsNo = static_cast<uint32>(std::distance(m_Instruments.data(), &dlsIns));
-					// cppcheck false-positive
-					// cppcheck-suppress returnDanglingLifetime
 					return &dlsIns;
+				} else if(region.uKeyMin > key)
+				{
+					// Regions are sorted, if we arrived here we won't find anything in the remaining regions
+					break;
 				}
 			}
+		} else
+		{
+			if(pInsNo)
+				*pInsNo = static_cast<uint32>(std::distance(m_Instruments.data(), &dlsIns));
+			return &dlsIns;
 		}
 	}
 
@@ -779,13 +876,14 @@ bool CDLSBank::UpdateInstrumentDefinition(DLSINSTRUMENT *pDlsIns, FileReader chu
 				if(!(pDlsIns->ulBank & F_INSTRUMENT_DRUMS))
 				{
 					pDlsIns->nMelodicEnv = static_cast<uint32>(m_Envelopes.size() + 1);
+				} else
+				{
+					if(!pDlsIns->Regions.empty())
+						pDlsIns->Regions.back().uPercEnv = static_cast<uint32>(m_Envelopes.size() + 1);
 				}
 				if(art1.cbSize + art1.cConnectionBlocks * sizeof(ConnectionBlock) > header.len)
 					break;
 				DLSENVELOPE dlsEnv;
-				MemsetZero(dlsEnv);
-				dlsEnv.nDefPan = 128;
-				dlsEnv.nVolSustainLevel = 128;
 				//Log("  art1 (%3d bytes): cbSize=%d cConnectionBlocks=%d\n", p->len, p->cbSize, p->cConnectionBlocks);
 				chunk.Seek(sizeof(IFFCHUNK) + art1.cbSize);
 				for (uint32 iblk = 0; iblk < art1.cConnectionBlocks; iblk++)
@@ -802,57 +900,85 @@ bool CDLSBank::UpdateInstrumentDefinition(DLSINSTRUMENT *pDlsIns, FileReader chu
 					case ART_DEFAULTPAN:
 						{
 							int32 pan = 128 + blk.lScale / (65536000/128);
-							dlsEnv.nDefPan = mpt::saturate_cast<uint8>(pan);
+							dlsEnv.defaultPan = mpt::saturate_cast<uint8>(pan);
 						}
 						break;
 
+					case ART_VOL_EG_DELAYTIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						dlsEnv.volumeEnv.delay = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
+						break;
+
 					case ART_VOL_EG_ATTACKTIME:
 						// 32-bit time cents units. range = [0s, 20s]
-						dlsEnv.wVolAttack = 0;
 						if(blk.lScale > -0x40000000)
 						{
-							int32 l = blk.lScale - 78743200; // maximum velocity
-							if (l > 0) l = 0;
-							int32 attacktime = DLS32BitTimeCentsToMilliseconds(l);
-							if (attacktime < 0) attacktime = 0;
-							if (attacktime > 20000) attacktime = 20000;
-							if (attacktime >= 20) dlsEnv.wVolAttack = (uint16)(attacktime / 20);
-							//Log("%3d: Envelope Attack Time set to %d (%d time cents)\n", (uint32)(dlsEnv.ulInstrument & 0x7F)|((dlsEnv.ulBank >> 16) & 0x8000), attacktime, pblk->lScale);
+							int32 l = std::min(0, blk.lScale - 78743200); // maximum velocity
+							dlsEnv.volumeEnv.attack = DLSEnvelopeTimeCentsToMilliseconds(l);
 						}
 						break;
 
+					case ART_VOL_EG_HOLDTIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						dlsEnv.volumeEnv.hold = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
+						break;
+
 					case ART_VOL_EG_DECAYTIME:
 						// 32-bit time cents units. range = [0s, 20s]
-						dlsEnv.wVolDecay = 0;
-						if(blk.lScale > -0x40000000)
-						{
-							int32 decaytime = DLS32BitTimeCentsToMilliseconds(blk.lScale);
-							if (decaytime > 20000) decaytime = 20000;
-							if (decaytime >= 20) dlsEnv.wVolDecay = (uint16)(decaytime / 20);
-							//Log("%3d: Envelope Decay Time set to %d (%d time cents)\n", (uint32)(dlsEnv.ulInstrument & 0x7F)|((dlsEnv.ulBank >> 16) & 0x8000), decaytime, pblk->lScale);
-						}
+						dlsEnv.volumeEnv.decay = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
 						break;
 
 					case ART_VOL_EG_RELEASETIME:
 						// 32-bit time cents units. range = [0s, 20s]
-						dlsEnv.wVolRelease = 0;
-						if(blk.lScale > -0x40000000)
-						{
-							int32 releasetime = DLS32BitTimeCentsToMilliseconds(blk.lScale);
-							if (releasetime > 20000) releasetime = 20000;
-							if (releasetime >= 20) dlsEnv.wVolRelease = (uint16)(releasetime / 20);
-							//Log("%3d: Envelope Release Time set to %d (%d time cents)\n", (uint32)(dlsEnv.ulInstrument & 0x7F)|((dlsEnv.ulBank >> 16) & 0x8000), dlsEnv.wVolRelease, pblk->lScale);
-						}
+						dlsEnv.volumeEnv.release = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
 						break;
 
 					case ART_VOL_EG_SUSTAINLEVEL:
 						// 0.1% units
 						if(blk.lScale >= 0)
 						{
-							dlsEnv.nVolSustainLevel = DLSSustainLevelToLinear(blk.lScale);
+							dlsEnv.volumeEnv.sustainLevel = DLSSustainLevelToLinear(blk.lScale);
 						}
 						break;
 
+					case ART_PITCH_EG_DELAYTIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						dlsEnv.pitchEnv.delay = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
+						break;
+
+					case ART_PITCH_EG_ATTACKTIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						if(blk.lScale > -0x40000000)
+						{
+							int32 l = std::min(0, blk.lScale - 78743200);  // maximum velocity
+							dlsEnv.pitchEnv.attack = DLSEnvelopeTimeCentsToMilliseconds(l);
+						}
+						break;
+
+					case ART_PITCH_EG_HOLDTIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						dlsEnv.pitchEnv.hold = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
+						break;
+
+					case ART_PITCH_EG_DECAYTIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						dlsEnv.pitchEnv.decay = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
+						break;
+
+					case ART_PITCH_EG_RELEASETIME:
+						// 32-bit time cents units. range = [0s, 20s]
+						dlsEnv.pitchEnv.release = DLSEnvelopeTimeCentsToMilliseconds(blk.lScale);
+						break;
+
+					case ART_PITCH_EG_SUSTAINLEVEL:
+						// 0.1% units
+						dlsEnv.pitchEnv.sustainLevel = DLSSustainLevelToLinear(blk.lScale);
+						break;
+
+					case ART_PITCH_EG_DEPTH:
+						dlsEnv.pitchEnvDepth = mpt::saturate_cast<int16>(blk.lScale / 65536);
+						break;
+
 					//default:
 					//	Log("    Articulation = 0x%08X value=%d\n", dwArticulation, blk.lScale);
 					}
@@ -864,15 +990,15 @@ bool CDLSBank::UpdateInstrumentDefinition(DLSINSTRUMENT *pDlsIns, FileReader chu
 		case IFFID_INAM:
 			chunk.ReadString<mpt::String::spacePadded>(pDlsIns->szName, header.len);
 			break;
-	#if 0
 		default:
-			{
-				char sid[5];
+#ifdef DLSINSTR_LOG
+		{
+				char sid[5]{};
 				memcpy(sid, &header.id, 4);
-				sid[4] = 0;
-				Log("    \"%s\": %d bytes\n", (uint32)sid, header.len.get());
+				MPT_LOG_GLOBAL(LogDebug, "DLSINSTR", MPT_UFORMAT("Unsupported DLS chunk: {} ({} bytes)")(mpt::ToUnicode(mpt::Charset::ASCII, mpt::String::ReadAutoBuf(sid)), header.len.get()));
 			}
-	#endif
+#endif
+			break;
 		}
 	}
 	return true;
@@ -1010,27 +1136,20 @@ bool CDLSBank::UpdateSF2PresetData(SF2LoaderInfo &sf2info, const IFFCHUNK &heade
 		}
 		break;
 
-	#ifdef DLSINSTR_LOG
 	default:
+#ifdef DLSINSTR_LOG
 		{
-			char sdbg[5];
+			char sdbg[5]{};
 			memcpy(sdbg, &header.id, 4);
-			sdbg[4] = 0;
 			MPT_LOG_GLOBAL(LogDebug, "DLSINSTR", MPT_UFORMAT("Unsupported SF2 chunk: {} ({} bytes)")(mpt::ToUnicode(mpt::Charset::ASCII, mpt::String::ReadAutoBuf(sdbg)), header.len.get()));
 		}
-	#endif
+#endif
+		break;
 	}
 	return true;
 }
 
 
-static int16 SF2TimeToDLS(int16 amount)
-{
-	int32 time = CDLSBank::DLS32BitTimeCentsToMilliseconds(static_cast<int32>(amount) << 16);
-	return static_cast<int16>(Clamp(time, 20, 20000) / 20);
-}
-
-
 // Convert all instruments to the DLS format
 bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 {
@@ -1049,12 +1168,6 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 		DLSENVELOPE dlsEnv;
 		int32 instrAttenuation = 0;
 		int16 instrFinetune = 0;
-		// Default Envelope Values
-		dlsEnv.wVolAttack = 0;
-		dlsEnv.wVolDecay = 0;
-		dlsEnv.wVolRelease = 0;
-		dlsEnv.nVolSustainLevel = 128;
-		dlsEnv.nDefPan = 128;
 		// Load Preset Bags
 		sf2info.presetBags.Seek(dlsIns.wPresetBagNdx * sizeof(SFPresetBag));
 		for(uint32 ipbagcnt = 0; ipbagcnt < dlsIns.wPresetBagNum; ipbagcnt++)
@@ -1074,22 +1187,6 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 				const int16 value = static_cast<int16>(gen.genAmount);
 				switch(gen.sfGenOper)
 				{
-				case SF2_GEN_ATTACKVOLENV:
-					dlsEnv.wVolAttack = SF2TimeToDLS(gen.genAmount);
-					break;
-				case SF2_GEN_DECAYVOLENV:
-					dlsEnv.wVolDecay = SF2TimeToDLS(gen.genAmount);
-					break;
-				case SF2_GEN_SUSTAINVOLENV:
-					// 0.1% units
-					if(gen.genAmount >= 0)
-					{
-						dlsEnv.nVolSustainLevel = SF2SustainLevelToLinear(gen.genAmount);
-					}
-					break;
-				case SF2_GEN_RELEASEVOLENV:
-					dlsEnv.wVolRelease = SF2TimeToDLS(gen.genAmount);
-					break;
 				case SF2_GEN_INSTRUMENT:
 					if(const auto instr = std::make_pair(gen.genAmount.get(), keyRange); !mpt::contains(instruments, instr))
 						instruments.push_back(instr);
@@ -1107,12 +1204,15 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 				case SF2_GEN_FINETUNE:
 					instrFinetune += static_cast<int16>(Util::muldiv(static_cast<int8>(value), 128, 100));
 					break;
-#if 0
 				default:
-					Log("Ins %3d: bag %3d gen %3d: ", nIns, ipbagndx, ipgenndx);
-					Log("genoper=%d amount=0x%04X ", gen.sfGenOper, gen.genAmount);
-					Log((pSmp->ulBank & F_INSTRUMENT_DRUMS) ? "(drum)\n" : "\n");
+					if(!gen.ApplyToEnvelope(dlsEnv))
+					{
+#ifdef DLSINSTR_LOG
+						MPT_LOG_GLOBAL(LogDebug, "DLSINSTR", MPT_UFORMAT("Preset {} bag {} gen {}: genoper={} amount={}{}")
+							(static_cast<int>(&dlsIns - m_Instruments.data()), ipbagcnt, static_cast<int>(&gen - generators.data()), gen.sfGenOper, gen.genAmount, (dlsIns.ulBank & F_INSTRUMENT_DRUMS) ? U_(" (drum)") : U_("")));
 #endif
+					}
+					break;
 				}
 			}
 		}
@@ -1164,6 +1264,13 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 
 				DLSREGION rgn = globalZone;
 
+				if(dlsIns.ulBank & F_INSTRUMENT_DRUMS)
+				{
+					m_Envelopes.push_back(dlsEnv);
+					rgn.uPercEnv = static_cast<uint32>(m_Envelopes.size());
+					pDlsEnv = &m_Envelopes[rgn.uPercEnv - 1];
+				}
+
 				// Region Default Values
 				int32 regionAttn = 0;
 				// Load Generators
@@ -1198,28 +1305,12 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 					case SF2_GEN_UNITYNOTE:
 						if (value < 128) rgn.uUnityNote = static_cast<uint8>(value);
 						break;
-					case SF2_GEN_ATTACKVOLENV:
-						pDlsEnv->wVolAttack = SF2TimeToDLS(gen.genAmount);
-						break;
-					case SF2_GEN_DECAYVOLENV:
-						pDlsEnv->wVolDecay = SF2TimeToDLS(gen.genAmount);
-						break;
-					case SF2_GEN_SUSTAINVOLENV:
-						// 0.1% units
-						if(gen.genAmount >= 0)
-						{
-							pDlsEnv->nVolSustainLevel = SF2SustainLevelToLinear(gen.genAmount);
-						}
-						break;
-					case SF2_GEN_RELEASEVOLENV:
-						pDlsEnv->wVolRelease = SF2TimeToDLS(gen.genAmount);
-						break;
 					case SF2_GEN_PAN:
 						{
 							int32 pan = static_cast<int16>(value);
 							pan = std::clamp(Util::muldivr(pan + 500, 256, 1000), 0, 256);
 							rgn.panning = static_cast<int16>(pan);
-							pDlsEnv->nDefPan = mpt::saturate_cast<uint8>(pan);
+							pDlsEnv->defaultPan = mpt::saturate_cast<uint8>(pan);
 						}
 						break;
 					case SF2_GEN_ATTENUATION:
@@ -1268,8 +1359,15 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 					case SF2_GEN_END_LOOP_COARSE:
 						loopEnd += static_cast<int16>(value) * 32768;
 						break;
-					//default:
-					//	Log("    gen=%d value=%04X\n", pgen->sfGenOper, pgen->genAmount);
+					default:
+						if(!gen.ApplyToEnvelope(*pDlsEnv))
+						{
+#ifdef DLSINSTR_LOG
+							MPT_LOG_GLOBAL(LogDebug, "DLSINSTR", MPT_UFORMAT("Instr {} region {} gen {}: genoper={} amount={}{}")
+								(nInstrNdx, nRgn, static_cast<int>(&gen - instrGenerators.data()), gen.sfGenOper, gen.genAmount, (dlsIns.ulBank & F_INSTRUMENT_DRUMS) ? U_(" (drum)") : U_("")));
+#endif
+						}
+						break;
 					}
 				}
 				int32 linearVol = DLS32BitRelativeGainToLinear(((instrAttenuation + regionAttn) * 65536) / 10) / 256;
@@ -1280,7 +1378,6 @@ bool CDLSBank::ConvertSF2ToDLS(SF2LoaderInfo &sf2info)
 					globalZone = rgn;
 				else if(!rgn.IsDummy())
 					dlsIns.Regions.push_back(rgn);
-				//Log("\n");
 			}
 		}
 	}
@@ -1295,7 +1392,7 @@ bool CDLSBank::Open(const mpt::PathString &filename)
 {
 	if(filename.empty()) return false;
 	m_szFileName = filename;
-	InputFile f(filename, SettingCacheCompleteFileBeforeLoading());
+	mpt::IO::InputFile f(filename, SettingCacheCompleteFileBeforeLoading());
 	if(!f.IsValid()) return false;
 	return Open(GetFileReader(f));
 }
@@ -1306,13 +1403,7 @@ bool CDLSBank::Open(FileReader file)
 	uint32 nInsDef;
 
 	if(file.GetOptionalFileName())
-	{
-#if defined(MPT_LIBCXX_QUIRK_NO_OPTIONAL_VALUE)
-		m_szFileName = *(file.GetOptionalFileName());
-#else
 		m_szFileName = file.GetOptionalFileName().value();
-#endif
-	}
 
 	file.Rewind();
 	size_t dwMemLength = file.GetLength();
@@ -1555,9 +1646,17 @@ bool CDLSBank::Open(FileReader file)
 	{
 		ConvertSF2ToDLS(sf2info);
 	}
-#ifdef DLSBANK_LOG
-	MPT_LOG_GLOBAL(LogDebug, "DLSBANK", U_("DLS bank closed"));
-#endif
+
+	// FindInstrument requires the instrument to be sorted for picking the best instrument from the MIDI library when there are multiple banks.
+	// And of course this is also helpful for creating the treeview UI
+	std::sort(m_Instruments.begin(), m_Instruments.end());
+	// Sort regions (for drums)
+	for(auto &instr : m_Instruments)
+	{
+		std::sort(instr.Regions.begin(), instr.Regions.end(), [](const DLSREGION &l, const DLSREGION &r)
+				  { return std::tie(l.uKeyMin, l.uKeyMax) < std::tie(r.uKeyMin, r.uKeyMax); });
+	}
+
 	return true;
 }
 
@@ -1572,7 +1671,10 @@ uint32 CDLSBank::GetRegionFromKey(uint32 nIns, uint32 nKey) const
 	for(uint32 rgn = 0; rgn < static_cast<uint32>(dlsIns.Regions.size()); rgn++)
 	{
 		const auto &region = dlsIns.Regions[rgn];
-		if(nKey < region.uKeyMin || nKey > region.uKeyMax)
+		// Regions are sorted, if we arrived here we won't find anything in the remaining regions
+		if(region.uKeyMin > nKey)
+			break;
+		if(nKey > region.uKeyMax)
 			continue;
 		if(region.nWaveLink == Util::MaxValueOfType(region.nWaveLink))
 			continue;
@@ -1649,7 +1751,7 @@ bool CDLSBank::ExtractWaveForm(uint32 nIns, uint32 nRgn, std::vector<uint8> &wav
 					{
 						waveData.assign(chunk.len + sizeof(IFFCHUNK), 0);
 						memcpy(waveData.data(), &chunk, sizeof(chunk));
-						mpt::IO::ReadRaw(f, &waveData[sizeof(chunk)], length - sizeof(chunk));
+						mpt::IO::ReadRaw(f, waveData.data() + sizeof(chunk), length - sizeof(chunk));
 					} catch(mpt::out_of_memory e)
 					{
 						mpt::delete_out_of_memory(e);
@@ -1806,6 +1908,129 @@ static uint16 ScaleEnvelope(uint32 time, float tempoScale)
 }
 
 
+uint32 DLSENVELOPE::Envelope::ConvertToMPT(InstrumentEnvelope &mptEnv, const EnvelopeType envType, const float tempoScale, const int16 valueScale) const
+{
+	if(!attack && decay >= 20 * 50 && (!sustainLevel && envType != ENV_PITCH) && release >= 20 * 50)
+		return uint32_max;
+
+	const EnvelopeNode::value_t neutralValue = (envType == ENV_VOLUME) ? 0 : ENVELOPE_MID;
+	
+	mptEnv.dwFlags.set(ENV_ENABLED);
+	mptEnv.clear();
+	// Delay section
+	uint16 attackStart = 0;
+	if(delay)
+	{
+		mptEnv.push_back(0, neutralValue);
+		attackStart = ScaleEnvelope(delay, tempoScale);
+	}
+	// Attack section
+	const auto attackValue = mpt::saturate_cast<EnvelopeNode::value_t>(neutralValue + Util::muldivr(ENVELOPE_MAX, valueScale, 3200));
+	if(attack)
+	{
+		mptEnv.push_back(attackStart, static_cast<EnvelopeNode::value_t>(attackValue / (attack / 2 + 2) + 8));  // /-----
+		mptEnv.push_back(attackStart + ScaleEnvelope(attack, tempoScale), attackValue);                         // |
+	} else
+	{
+		mptEnv.push_back(attackStart, attackValue);
+	}
+	// Hold section
+	if(EnvelopeNode::tick_t holdTick = attackStart + ScaleEnvelope(attack + hold, tempoScale); hold > 0 && holdTick > mptEnv.back().tick)
+		mptEnv.push_back(holdTick, attackValue);
+	// Sustain Level
+	if(sustainLevel > 0 || envType == ENV_PITCH)
+	{
+		if(sustainLevel < 128)
+		{
+			uint16 lStartTime = mptEnv.back().tick;
+			int32 lSusLevel = -CDLSBank::DLS32BitRelativeLinearToGain(sustainLevel << 9) / 65536;
+			int32 lDecayTime = 1;
+			if(lSusLevel > 0)
+			{
+				lDecayTime = (lSusLevel * (int32)decay) / 960;
+				for(uint32 i = 0; i < 7; i++)
+				{
+					int32 lFactor = 128 - (1 << i);
+					if(lFactor <= sustainLevel) break;
+					int32 lev = -CDLSBank::DLS32BitRelativeLinearToGain(lFactor << 9) / 65536;
+					if(lev > 0)
+					{
+						int32 ltime = (lev * (int32)decay) / 960;
+						if((ltime > 1) && (ltime < lDecayTime))
+						{
+							uint16 tick = lStartTime + ScaleEnvelope(ltime, tempoScale);
+							if(tick > mptEnv.back().tick)
+							{
+								mptEnv.push_back(tick, mpt::saturate_cast<EnvelopeNode::value_t>(neutralValue + Util::muldivr(lFactor, valueScale, 6400)));
+							}
+						}
+					}
+				}
+			}
+
+			uint16 decayEnd = lStartTime + ScaleEnvelope(lDecayTime, tempoScale);
+			if(decayEnd > mptEnv.back().tick)
+			{
+				mptEnv.push_back(decayEnd, mpt::saturate_cast<EnvelopeNode::value_t>(neutralValue + Util::muldivr(sustainLevel + 1, valueScale, 6400)));
+			}
+		}
+		mptEnv.dwFlags.set(ENV_SUSTAIN);
+	} else
+	{
+		mptEnv.dwFlags.set(ENV_SUSTAIN);
+		mptEnv.push_back(mptEnv.back().tick + 1u, mptEnv.back().value);
+	}
+	mptEnv.nSustainStart = mptEnv.nSustainEnd = (uint8)(mptEnv.size() - 1);
+	if(mptEnv.nSustainEnd > 0 && envType == ENV_VOLUME)
+		mptEnv.nReleaseNode = mptEnv.nSustainEnd;
+	// Release section
+	const bool lastIsNeutral = mptEnv.back().value > 1 && mptEnv.back().value == neutralValue;
+	if(release && mptEnv.back().value > 1 && !lastIsNeutral)
+	{
+		int32 lReleaseTime = release;
+		uint16 lStartTime = mptEnv.back().tick;
+		int32 lStartFactor = mptEnv.back().value;
+		int32 lSusLevel = -CDLSBank::DLS32BitRelativeLinearToGain(lStartFactor << 10) / 65536;
+		int32 lDecayEndTime = (lReleaseTime * lSusLevel) / 960;
+		lReleaseTime -= lDecayEndTime;
+		int32 prevFactor = lStartFactor;
+		for(uint32 i = 0; i < 7; i++)
+		{
+			int32 lFactor = 1 + ((lStartFactor * 3) >> (i + 2));
+			if((lFactor < 1) ||(lFactor == 1 && prevFactor == 1) || (lFactor >= lStartFactor))
+				continue;
+			prevFactor = lFactor;
+			int32 lev = -CDLSBank::DLS32BitRelativeLinearToGain(lFactor << 10) / 65536;
+			if(lev > 0)
+			{
+				int32 ltime = (((int32)release * lev) / 960) - lDecayEndTime;
+				if((ltime > 1) && (ltime < lReleaseTime))
+				{
+					uint16 tick = lStartTime + ScaleEnvelope(ltime, tempoScale);
+					if(tick > mptEnv.back().tick)
+					{
+						mptEnv.push_back(tick, mpt::saturate_cast<EnvelopeNode::value_t>(neutralValue + Util::muldivr(lFactor, valueScale, 3200)));
+						if(mptEnv.back().value ==  neutralValue)
+							break;
+					}
+				}
+			}
+		}
+		if(lReleaseTime < 1) lReleaseTime = 1;
+		auto releaseTicks = ScaleEnvelope(lReleaseTime, tempoScale);
+		mptEnv.push_back(lStartTime + releaseTicks, neutralValue);
+		if(releaseTicks > 0)
+		{
+			return 32768 / releaseTicks;
+		}
+	} else if(!lastIsNeutral)
+	{
+		mptEnv.push_back(mptEnv.back().tick + 1u, neutralValue);
+	}
+	return uint32_max;
+}
+
+
 bool CDLSBank::ExtractInstrument(CSoundFile &sndFile, INSTRUMENTINDEX nInstr, uint32 nIns, uint32 nDrumRgn) const
 {
 	uint32 minRegion, maxRegion, nEnv;
@@ -2038,109 +2263,10 @@ bool CDLSBank::ExtractInstrument(CSoundFile &sndFile, INSTRUMENTINDEX nInstr, ui
 	if ((nEnv) && (nEnv <= m_Envelopes.size()))
 	{
 		const DLSENVELOPE &part = m_Envelopes[nEnv - 1];
-		// Volume Envelope
-		if ((part.wVolAttack) || (part.wVolDecay < 20*50) || (part.nVolSustainLevel) || (part.wVolRelease < 20*50))
-		{
-			pIns->VolEnv.dwFlags.set(ENV_ENABLED);
-			// Delay section
-			// -> DLS level 2
-			// Attack section
-			pIns->VolEnv.clear();
-			if (part.wVolAttack)
-			{
-				pIns->VolEnv.push_back(0, (uint8)(ENVELOPE_MAX / (part.wVolAttack / 2 + 2) + 8)); // /-----
-				pIns->VolEnv.push_back(ScaleEnvelope(part.wVolAttack, tempoScale), ENVELOPE_MAX); // |
-			} else
-			{
-				pIns->VolEnv.push_back(0, ENVELOPE_MAX);
-			}
-			// Hold section
-			// -> DLS Level 2
-			// Sustain Level
-			if (part.nVolSustainLevel > 0)
-			{
-				if (part.nVolSustainLevel < 128)
-				{
-					uint16 lStartTime = pIns->VolEnv.back().tick;
-					int32 lSusLevel = - DLS32BitRelativeLinearToGain(part.nVolSustainLevel << 9) / 65536;
-					int32 lDecayTime = 1;
-					if (lSusLevel > 0)
-					{
-						lDecayTime = (lSusLevel * (int32)part.wVolDecay) / 960;
-						for (uint32 i=0; i<7; i++)
-						{
-							int32 lFactor = 128 - (1 << i);
-							if (lFactor <= part.nVolSustainLevel) break;
-							int32 lev = - DLS32BitRelativeLinearToGain(lFactor << 9) / 65536;
-							if (lev > 0)
-							{
-								int32 ltime = (lev * (int32)part.wVolDecay) / 960;
-								if ((ltime > 1) && (ltime < lDecayTime))
-								{
-									uint16 tick = lStartTime + ScaleEnvelope(ltime, tempoScale);
-									if(tick > pIns->VolEnv.back().tick)
-									{
-										pIns->VolEnv.push_back(tick, (uint8)(lFactor / 2));
-									}
-								}
-							}
-						}
-					}
-
-					uint16 decayEnd = lStartTime + ScaleEnvelope(lDecayTime, tempoScale);
-					if (decayEnd > pIns->VolEnv.back().tick)
-					{
-						pIns->VolEnv.push_back(decayEnd, (uint8)((part.nVolSustainLevel+1) / 2));
-					}
-				}
-				pIns->VolEnv.dwFlags.set(ENV_SUSTAIN);
-			} else
-			{
-				pIns->VolEnv.dwFlags.set(ENV_SUSTAIN);
-				pIns->VolEnv.push_back(pIns->VolEnv.back().tick + 1u, pIns->VolEnv.back().value);
-			}
-			pIns->VolEnv.nSustainStart = pIns->VolEnv.nSustainEnd = (uint8)(pIns->VolEnv.size() - 1);
-			// Release section
-			if ((part.wVolRelease) && (pIns->VolEnv.back().value > 1))
-			{
-				int32 lReleaseTime = part.wVolRelease;
-				uint16 lStartTime = pIns->VolEnv.back().tick;
-				int32 lStartFactor = pIns->VolEnv.back().value;
-				int32 lSusLevel = - DLS32BitRelativeLinearToGain(lStartFactor << 10) / 65536;
-				int32 lDecayEndTime = (lReleaseTime * lSusLevel) / 960;
-				lReleaseTime -= lDecayEndTime;
-				if(pIns->VolEnv.nSustainEnd > 0)
-					pIns->VolEnv.nReleaseNode = pIns->VolEnv.nSustainEnd;
-				for (uint32 i=0; i<5; i++)
-				{
-					int32 lFactor = 1 + ((lStartFactor * 3) >> (i+2));
-					if ((lFactor <= 1) || (lFactor >= lStartFactor)) continue;
-					int32 lev = - DLS32BitRelativeLinearToGain(lFactor << 10) / 65536;
-					if (lev > 0)
-					{
-						int32 ltime = (((int32)part.wVolRelease * lev) / 960) - lDecayEndTime;
-						if ((ltime > 1) && (ltime < lReleaseTime))
-						{
-							uint16 tick = lStartTime + ScaleEnvelope(ltime, tempoScale);
-							if(tick > pIns->VolEnv.back().tick)
-							{
-								pIns->VolEnv.push_back(tick, (uint8)lFactor);
-							}
-						}
-					}
-				}
-				if (lReleaseTime < 1) lReleaseTime = 1;
-				auto releaseTicks = ScaleEnvelope(lReleaseTime, tempoScale);
-				pIns->VolEnv.push_back(lStartTime + releaseTicks, ENVELOPE_MIN);
-				if(releaseTicks > 0)
-				{
-					pIns->nFadeOut = 32768 / releaseTicks;
-				}
-			} else
-			{
-				pIns->VolEnv.push_back(pIns->VolEnv.back().tick + 1u, ENVELOPE_MIN);
-			}
-		}
+		if(const auto fadeout = part.volumeEnv.ConvertToMPT(pIns->VolEnv, ENV_VOLUME, tempoScale, 3200); fadeout != uint32_max)
+			pIns->nFadeOut = fadeout;
+		if(std::abs(part.pitchEnvDepth) >= 50)
+			part.pitchEnv.ConvertToMPT(pIns->PitchEnv, ENV_PITCH, tempoScale, part.pitchEnvDepth);
 	}
 	if(isDrum)
 	{
@@ -2195,13 +2321,13 @@ uint16 CDLSBank::GetPanning(uint32 ins, uint32 region) const
 	{
 		if(rgn.uPercEnv > 0 && rgn.uPercEnv <= m_Envelopes.size())
 		{
-			return m_Envelopes[rgn.uPercEnv - 1].nDefPan;
+			return m_Envelopes[rgn.uPercEnv - 1].defaultPan;
 		}
 	} else
 	{
 		if(dlsIns.nMelodicEnv > 0 && dlsIns.nMelodicEnv <= m_Envelopes.size())
 		{
-			return m_Envelopes[dlsIns.nMelodicEnv - 1].nDefPan;
+			return m_Envelopes[dlsIns.nMelodicEnv - 1].defaultPan;
 		}
 	}
 	return 128;
