@@ -4,7 +4,7 @@
  * Purpose: IMF (Imago Orpheus) module loader
  * Notes  : Reverb and Chorus are not supported.
  * Authors: Storlek (Original author - http://schismtracker.org/ - code ported with permission)
- *			Johannes Schultz (OpenMPT Port, tweaks)
+ *          Johannes Schultz (OpenMPT Port, tweaks)
  * The OpenMPT source code is released under the BSD license. Read LICENSE for more details.
  */
 
@@ -111,7 +111,7 @@ struct IMFInstrument
 		uint16 minTick = 0; // minimum tick value for next node
 		for(uint32 n = 0; n < mptEnv.size(); n++)
 		{
-			minTick = mptEnv[n].tick = std::max<uint16>(minTick, nodes[e][n].tick);
+			mptEnv[n].tick = minTick = std::max(minTick, nodes[e][n].tick.get());
 			minTick++;
 			uint8 value = static_cast<uint8>(nodes[e][n].value ^ mirror) >> shift;
 			mptEnv[n].value = std::min(value, uint8(ENVELOPE_MAX));
@@ -122,17 +122,18 @@ struct IMFInstrument
 	// Convert an IMFInstrument to OpenMPT's internal instrument representation.
 	void ConvertToMPT(ModInstrument &mptIns, SAMPLEINDEX firstSample) const
 	{
-		mpt::String::Read<mpt::String::nullTerminated>(mptIns.name, name);
+		mptIns.name = mpt::String::ReadBuf(mpt::String::nullTerminated, name);
 
 		if(smpNum)
 		{
-			for(size_t note = 0; note < std::min(CountOf(map), CountOf(mptIns.Keyboard) - 12u); note++)
+			for(size_t note = 0; note < std::min(std::size(map), std::size(mptIns.Keyboard) - 12u); note++)
 			{
 				mptIns.Keyboard[note + 12] = firstSample + map[note];
 			}
 		}
 
 		mptIns.nFadeOut = fadeout;
+		mptIns.midiPWD = 1;  // For CMD_FINETUNE
 
 		ConvertEnvelope(mptIns.VolEnv, volEnv);
 		ConvertEnvelope(mptIns.PanEnv, panEnv);
@@ -177,7 +178,7 @@ struct IMFSample
 	void ConvertToMPT(ModSample &mptSmp) const
 	{
 		mptSmp.Initialize();
-		mpt::String::Read<mpt::String::nullTerminated>(mptSmp.filename, filename);
+		mptSmp.filename = mpt::String::ReadBuf(mpt::String::nullTerminated, filename);
 
 		mptSmp.nLength = length;
 		mptSmp.nLoopStart = loopStart;
@@ -204,7 +205,7 @@ struct IMFSample
 MPT_BINARY_STRUCT(IMFSample, 64)
 
 
-static const EffectCommand imfEffects[] =
+static constexpr EffectCommand imfEffects[] =
 {
 	CMD_NONE,
 	CMD_SPEED,           // 0x01 1xx Set Tempo
@@ -221,7 +222,7 @@ static const EffectCommand imfEffects[] =
 	CMD_VOLUME,          // 0x0C Cxx Set Volume
 	CMD_VOLUMESLIDE,     // 0x0D Dxy Volume Slide
 	CMD_VOLUMESLIDE,     // 0x0E Exy Fine Volume Slide
-	CMD_S3MCMDEX,        // 0x0F Fxx Set Finetune
+	CMD_FINETUNE,        // 0x0F Fxx Set Finetune
 	CMD_NOTESLIDEUP,     // 0x10 Gxy Note Slide Up
 	CMD_NOTESLIDEDOWN,   // 0x11 Hxy Note Slide Down
 	CMD_PORTAMENTOUP,    // 0x12 Ixx Slide Up
@@ -254,49 +255,48 @@ static const EffectCommand imfEffects[] =
 	CMD_NONE,            // 0x23 Zxx Reverb - XXX
 };
 
-static void ImportIMFEffect(ModCommand &m)
+static std::pair<EffectCommand, uint8> TranslateIMFEffect(uint8 command, uint8 param)
 {
 	uint8 n;
 	// fix some of them
-	switch(m.command)
+	switch(command)
 	{
 	case 0xE: // fine volslide
 		// hackaround to get almost-right behavior for fine slides (i think!)
-		if(m.param == 0)
+		if(param == 0)
 			/* nothing */;
-		else if(m.param == 0xF0)
-			m.param = 0xEF;
-		else if(m.param == 0x0F)
-			m.param = 0xFE;
-		else if(m.param & 0xF0)
-			m.param |= 0x0F;
+		else if(param == 0xF0)
+			param = 0xEF;
+		else if(param == 0x0F)
+			param = 0xFE;
+		else if(param & 0xF0)
+			param |= 0x0F;
 		else
-			m.param |= 0xF0;
+			param |= 0xF0;
 		break;
 	case 0xF: // set finetune
-		// we don't implement this, but let's at least import the value
-		m.param = 0x20 | MIN(m.param >> 4, 0x0F);
+		param ^= 0x80;
 		break;
 	case 0x14: // fine slide up
 	case 0x15: // fine slide down
 		// this is about as close as we can do...
-		if(m.param >> 4)
-			m.param = 0xF0 | (m.param >> 4);
+		if(param >> 4)
+			param = 0xF0 | (param >> 4);
 		else
-			m.param |= 0xE0;
+			param |= 0xE0;
 		break;
 	case 0x16: // cutoff
-		m.param = (0xFF - m.param) / 2u;
+		param = (0xFF - param) / 2u;
 		break;
 	case 0x17: // cutoff slide + resonance (TODO: cutoff slide is currently not handled)
-		m.param = 0x80 | (m.param & 0x0F);
+		param = 0x80 | (param & 0x0F);
 		break;
 	case 0x1F: // set global volume
-		m.param = MIN(m.param << 1, 0xFF);
+		param = mpt::saturate_cast<uint8>(param * 2);
 		break;
 	case 0x21:
 		n = 0;
-		switch (m.param >> 4)
+		switch(param >> 4)
 		{
 		case 0:
 			/* undefined, but since S0x does nothing in IT anyway, we won't care.
@@ -306,7 +306,7 @@ static void ImportIMFEffect(ModCommand &m)
 		default: // undefined
 		case 0x1: // set filter
 		case 0xF: // invert loop
-			m.command = CMD_NONE;
+			command = 0;
 			break;
 		case 0x3: // glissando
 			n = 0x20;
@@ -326,41 +326,34 @@ static void ImportIMFEffect(ModCommand &m)
 		case 0xC: // note cut
 		case 0xD: // note delay
 			// Apparently, Imago Orpheus doesn't cut samples on tick 0.
-			if(!m.param)
-				m.command = CMD_NONE;
+			if(!param)
+				command = 0;
 			break;
 		case 0xE: // ignore envelope
-			switch(m.param & 0x0F)
+			switch(param & 0x0F)
 			{
 			// All envelopes
 			// Predicament: we can only disable one envelope at a time. Volume is probably most noticeable, so let's go with that.
-			case 0: m.param = 0x77; break;
+			case 0: param = 0x77; break;
 			// Volume
-			case 1: m.param = 0x77; break;
+			case 1: param = 0x77; break;
 			// Panning
-			case 2: m.param = 0x79; break;
+			case 2: param = 0x79; break;
 			// Filter
-			case 3: m.param = 0x7B; break;
+			case 3: param = 0x7B; break;
 			}
 			break;
 		case 0x18: // sample offset
 			// O00 doesn't pick up the previous value
-			if(!m.param)
-				m.command = CMD_NONE;
+			if(!param)
+				command = 0;
 			break;
 		}
 		if(n)
-			m.param = n | (m.param & 0x0F);
+			param = n | (param & 0x0F);
 		break;
 	}
-	m.command = (m.command < CountOf(imfEffects)) ? imfEffects[m.command] : CMD_NONE;
-	if(m.command == CMD_VOLUME && m.volcmd == VOLCMD_NONE)
-	{
-		m.volcmd = VOLCMD_VOLUME;
-		m.vol = m.param;
-		m.command = CMD_NONE;
-		m.param = 0;
-	}
+	return {(command < std::size(imfEffects)) ? imfEffects[command] : CMD_NONE, param};
 }
 
 
@@ -404,8 +397,7 @@ static bool ValidateHeader(const IMFFileHeader &fileHeader)
 
 static uint64 GetHeaderMinimumAdditionalSize(const IMFFileHeader &fileHeader)
 {
-	MPT_UNREFERENCED_PARAMETER(fileHeader);
-	return 256;
+	return 256 + fileHeader.patNum * 4 + fileHeader.insNum * sizeof(IMFInstrument);
 }
 
 
@@ -440,6 +432,10 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 	{
 		return false;
 	}
+	if(loadFlags == onlyVerifyHeader)
+	{
+		return true;
+	}
 
 	// Read channel configuration
 	std::bitset<32> ignoreChannels; // bit set for each channel that's completely disabled
@@ -449,7 +445,7 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 		ChnSettings[chn].Reset();
 		ChnSettings[chn].nPan = fileHeader.channels[chn].panning * 256 / 255;
 
-		mpt::String::Read<mpt::String::nullTerminated>(ChnSettings[chn].szName, fileHeader.channels[chn].name);
+		ChnSettings[chn].szName = mpt::String::ReadBuf(mpt::String::nullTerminated, fileHeader.channels[chn].name);
 
 		// TODO: reverb/chorus?
 		switch(fileHeader.channels[chn].status)
@@ -466,23 +462,17 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 			ignoreChannels[chn] = true;
 			break;
 		default: // uhhhh.... freak out
-			//fprintf(stderr, "imf: channel %d has unknown status %d\n", n, hdr.channels[n].status);
 			return false;
 		}
 	}
-	if(!detectedChannels)
-	{
-		return false;
-	}
-	
-	if(loadFlags == onlyVerifyHeader)
-	{
-		return true;
-	}
 
 	InitializeGlobals(MOD_TYPE_IMF);
 	m_nChannels = detectedChannels;
 
+	m_modFormat.formatName = U_("Imago Orpheus");
+	m_modFormat.type = U_("imf");
+	m_modFormat.charset = mpt::Charset::CP437;
+
 	//From mikmod: work around an Orpheus bug
 	if(fileHeader.channels[0].status == 0)
 	{
@@ -496,9 +486,9 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 	}
 
 	// Song Name
-	mpt::String::Read<mpt::String::nullTerminated>(m_songName, fileHeader.title);
+	m_songName = mpt::String::ReadBuf(mpt::String::nullTerminated, fileHeader.title);
 
-	m_SongFlags = (fileHeader.flags & IMFFileHeader::linearSlides) ? SONG_LINEARSLIDES : SongFlags(0);
+	m_SongFlags.set(SONG_LINEARSLIDES, fileHeader.flags & IMFFileHeader::linearSlides);
 	m_nDefaultSpeed = fileHeader.tempo;
 	m_nDefaultTempo.Set(fileHeader.bpm);
 	m_nDefaultGlobalVolume = fileHeader.master * 4u;
@@ -524,7 +514,7 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 			continue;
 		}
 
-		ModCommand junkNote;
+		ModCommand dummy;
 		ROWINDEX row = 0;
 		while(row < numRows)
 		{
@@ -536,13 +526,14 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 			}
 
 			uint8 channel = mask & 0x1F;
-			ModCommand &m = ignoreChannels[channel] ? junkNote : *Patterns[pat].GetpModCommand(row, channel);
+			ModCommand &m = (channel < GetNumChannels()) ? *Patterns[pat].GetpModCommand(row, channel) : dummy;
 
 			if(mask & 0x20)
 			{
 				// Read note/instrument
-				m.note = patternChunk.ReadUint8();
-				m.instr = patternChunk.ReadUint8();
+				const auto [note, instr] = patternChunk.ReadArray<uint8, 2>();
+				m.note = note;
+				m.instr = instr;
 
 				if(m.note == 160)
 				{
@@ -562,51 +553,22 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 			if((mask & 0xC0) == 0xC0)
 			{
 				// Read both effects and figure out what to do with them
-				uint8 e1c = patternChunk.ReadUint8();	// Command 1
-				uint8 e1d = patternChunk.ReadUint8();	// Data 1
-				uint8 e2c = patternChunk.ReadUint8();	// Command 2
-				uint8 e2d = patternChunk.ReadUint8();	// Data 2
-
-				if(e1c == 0x0C)
-				{
-					m.vol = MIN(e1d, 0x40);
-					m.volcmd = VOLCMD_VOLUME;
-					m.command = e2c;
-					m.param = e2d;
-				} else if(e2c == 0x0C)
-				{
-					m.vol = MIN(e2d, 0x40);
-					m.volcmd = VOLCMD_VOLUME;
-					m.command = e1c;
-					m.param = e1d;
-				} else if(e1c == 0x0A)
-				{
-					m.vol = e1d * 64 / 255;
-					m.volcmd = VOLCMD_PANNING;
-					m.command = e2c;
-					m.param = e2d;
-				} else if(e2c == 0x0A)
-				{
-					m.vol = e2d * 64 / 255;
-					m.volcmd = VOLCMD_PANNING;
-					m.command = e1c;
-					m.param = e1d;
-				} else
-				{
-					/* check if one of the effects is a 'global' effect
-					-- if so, put it in some unused channel instead.
-					otherwise pick the most important effect. */
-					m.command = e2c;
-					m.param = e2d;
-				}
+				const auto [e1c, e1d, e2c, e2d] = patternChunk.ReadArray<uint8, 4>();  // Command 1, Data 1, Command 2, Data 2
+				const auto [command1, param1] = TranslateIMFEffect(e1c, e1d);
+				const auto [command2, param2] = TranslateIMFEffect(e2c, e2d);
+				m.FillInTwoCommands(command1, param1, command2, param2);
 			} else if(mask & 0xC0)
 			{
-				// There's one effect, just stick it in the effect column
-				m.command = patternChunk.ReadUint8();
-				m.param = patternChunk.ReadUint8();
+				// There's one effect, just stick it in the effect column (unless it's a volume command)
+				const auto [e1c, e1d] = patternChunk.ReadArray<uint8, 2>();  // Command 1, Data 1, Command 2, Data 2
+				const auto [command, param] = TranslateIMFEffect(e1c, e1d);
+				if(command == CMD_VOLUME)
+					m.SetVolumeCommand(VOLCMD_VOLUME, param);
+				else
+					m.SetEffectCommand(command, param);
 			}
-			if(m.command)
-				ImportIMFEffect(m);
+			if(ignoreChannels[channel] && m.IsGlobalCommand())
+				m.command = CMD_NONE;
 		}
 	}
 
@@ -643,7 +605,7 @@ bool CSoundFile::ReadIMF(FileReader &file, ModLoadingFlags loadFlags)
 			ModSample &sample = Samples[smpID];
 
 			sampleHeader.ConvertToMPT(sample);
-			mpt::String::Copy(m_szNames[smpID], sample.filename);
+			m_szNames[smpID] = sample.filename;
 
 			if(sampleHeader.length)
 			{
