@@ -10,10 +10,14 @@
 #include "stdafx.h"
 #include "Loaders.h"
 
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 #include "../mptrack/Vstplug.h"
 #include "plugins/PluginManager.h"
-#endif
+#endif  // MPT_WITH_VST
+#include "mpt/io/base.hpp"
+#include "mpt/io/io.hpp"
+#include "mpt/io/io_span.hpp"
+#include "mpt/io/io_stdstream.hpp"
 
 #include <map>
 
@@ -97,7 +101,7 @@ struct MMDSong
 	{
 		FLAG_FILTERON   = 0x01,  // The hardware audio filter is on
 		FLAG_JUMPINGON  = 0x02,  // Mouse pointer jumping on
-		FLAG_JUMP8TH    = 0x04,  // Jump every 8th line (not in OctaMED Pro)
+		FLAG_JUMP8TH    = 0x04,  // ump every 8th line (not in OctaMED Pro)
 		FLAG_INSTRSATT  = 0x08,  // sng+samples indicator (not useful in MMDs)
 		FLAG_VOLHEX     = 0x10,  // volumes are HEX
 		FLAG_STSLIDE    = 0x20,  // use ST/NT/PT compatible sliding
@@ -375,8 +379,9 @@ static TEMPO MMDTempoToBPM(uint32 tempo, bool is8Ch, bool bpmMode, uint8 rowsPer
 	if(is8Ch && tempo > 0)
 	{
 		LimitMax(tempo, 10u);
-		static constexpr uint8 tempos[10] = { 47, 43, 40, 37, 35, 32, 30, 29, 27, 26 };
-		tempo = tempos[tempo - 1];
+		// MED Soundstudio uses these tempos when importing old files
+		static constexpr uint8 tempos[10] = {179, 164, 152, 141, 131, 123, 116, 110, 104, 99};
+		return TEMPO(tempos[tempo - 1], 0);
 	} else if(tempo > 0 && tempo <= 10)
 	{
 		// SoundTracker compatible tempo
@@ -387,9 +392,10 @@ static TEMPO MMDTempoToBPM(uint32 tempo, bool is8Ch, bool bpmMode, uint8 rowsPer
 }
 
 
-static void ConvertMEDEffect(ModCommand &m, bool is8ch, bool bpmMode, uint8 rowsPerBeat, bool volHex)
+static std::pair<EffectCommand, ModCommand::PARAM> ConvertMEDEffect(ModCommand &m, const uint8 command, const bool is8ch, const bool bpmMode, const uint8 rowsPerBeat, const bool volHex)
 {
-	switch(m.command)
+	m.command = CMD_NONE;
+	switch(command)
 	{
 	case 0x04:  // Vibrato (twice as deep as in ProTracker)
 		m.command = CMD_VIBRATO;
@@ -399,7 +405,7 @@ static void ConvertMEDEffect(ModCommand &m, bool is8ch, bool bpmMode, uint8 rows
 		m.command = CMD_NONE;
 		break;
 	case 0x09:  // Set secondary speed
-		if(m.param > 0 && m.param <= 0x20)
+		if(m.param > 0 && m.param <= 20)
 			m.command = CMD_SPEED;
 		else
 			m.command = CMD_NONE;
@@ -426,11 +432,27 @@ static void ConvertMEDEffect(ModCommand &m, bool is8ch, bool bpmMode, uint8 rows
 		} else if(m.param <= 0xF0)
 		{
 			m.command = CMD_TEMPO;
-			if(m.param < 0x03)  // This appears to be a bug in OctaMED which is not emulated in MED Soundstudio on Windows.
+			if(m.param < 0x03)
+			{
+				// This appears to be a bug in OctaMED which is not emulated in MED Soundstudio on Windows.
 				m.param = 0x70;
-			else
-				m.param = mpt::saturate_round<ModCommand::PARAM>(MMDTempoToBPM(m.param, is8ch, bpmMode, rowsPerBeat).ToDouble());
-		} else switch(m.param)
+			} else
+			{
+				uint16 tempo = mpt::saturate_round<uint16>(MMDTempoToBPM(m.param, is8ch, bpmMode, rowsPerBeat).ToDouble());
+				if(tempo <= Util::MaxValueOfType(m.param))
+				{
+					m.param = static_cast<ModCommand::PARAM>(tempo);
+				} else
+				{
+					m.param = static_cast<ModCommand::PARAM>(tempo >> 8);
+					return {CMD_XPARAM, static_cast<ModCommand::PARAM>(tempo & 0xFF)};
+				}
+			}
+#ifdef MODPLUG_TRACKER
+			if(m.param < 0x20)
+				m.param = 0x20;
+#endif  // MODPLUG_TRACKER
+		} else switch(command)
 		{
 			case 0xF1:  // Play note twice
 				m.command = CMD_MODCMDEX;
@@ -456,8 +478,7 @@ static void ConvertMEDEffect(ModCommand &m, bool is8ch, bool bpmMode, uint8 rows
 				m.command = CMD_NONE;
 				break;
 			case 0xFF:  // Turn note off
-				if(!m.IsNote())
-					m.note = NOTE_NOTECUT;
+				m.note = NOTE_NOTECUT;
 				m.command = CMD_NONE;
 				break;
 			default:
@@ -572,18 +593,19 @@ static void ConvertMEDEffect(ModCommand &m, bool is8ch, bool bpmMode, uint8 rows
 		}
 		break;
 	default:
-		if(m.command < 0x10)
-			CSoundFile::ConvertModCommand(m);
+		if(command < 0x10)
+			CSoundFile::ConvertModCommand(m, command, m.param);
 		else
 			m.command = CMD_NONE;
 		break;
 	}
+	return std::make_pair(CMD_NONE, ModCommand::PARAM(0));
 }
 
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 static std::wstring ReadMEDStringUTF16BE(FileReader &file)
 {
-	auto chunk = file.ReadChunk(file.ReadUint32BE());
+	FileReader chunk = file.ReadChunk(file.ReadUint32BE());
 	std::wstring s(chunk.GetLength() / 2u, L'\0');
 	for(auto &c : s)
 	{
@@ -591,7 +613,7 @@ static std::wstring ReadMEDStringUTF16BE(FileReader &file)
 	}
 	return s;
 }
-#endif
+#endif  // MPT_WITH_VST
 
 
 static void MEDReadNextSong(FileReader &file, MMD0FileHeader &fileHeader, MMD0Exp &expData, MMDSong &songHeader)
@@ -739,13 +761,12 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 	// - starkelsesirap.mmd0 (synth instruments) on the other hand don't need it
 	// In MMD2 / MMD3, the mix flag is used instead.
 	const bool hardwareMixSamples = (version < 2) || (version >= 2 && !(songHeader.flags2 & MMDSong::FLAG2_MIX));
-	m_nMinPeriod = hardwareMixSamples ? (113 * 4) : (55 * 4);
 
 	bool needInstruments = false;
 	bool anySynthInstrs = false;
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 	PLUGINDEX numPlugins = 0;
-#endif
+#endif  // MPT_WITH_VST
 	for(SAMPLEINDEX ins = 1, smp = 1; ins <= m_nInstruments; ins++)
 	{
 		if(!AllocateInstrument(ins, smp))
@@ -765,7 +786,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 		const bool isSynth = instrHeader.type < 0;
 		const size_t maskedType = static_cast<size_t>(instrHeader.type & MMDInstrHeader::TYPEMASK);
 
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 		if(instrHeader.type == MMDInstrHeader::VSTI)
 		{
 			needInstruments = true;
@@ -778,7 +799,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 				mixPlug = {};
 				mixPlug.Info.dwPluginId1 = Vst::kEffectMagic;
 				mixPlug.Info.gain = 10;
-				mixPlug.Info.szName = mpt::ToCharset(mpt::CharsetLocaleOrUTF8, name);
+				mixPlug.Info.szName = mpt::ToCharset(mpt::Charset::Locale, name);
 				mixPlug.Info.szLibraryName = mpt::ToCharset(mpt::Charset::UTF8, name);
 				instr.nMixPlug = numPlugins + 1;
 				instr.nMidiChannel = MidiFirstChannel;
@@ -789,7 +810,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 				numPlugins++;
 			}
 		} else
-#endif  // NO_VST
+#endif  // MPT_WITH_VST
 		if(isSynth)
 		{
 			// TODO: Figure out synth instruments
@@ -797,9 +818,6 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 			instr.AssignSample(0);
 		}
 
-		MMD0Sample sampleHeader;
-		sampleHeaderChunk.ReadStruct(sampleHeader);
-
 		uint8 numSamples = 1;
 		static constexpr uint8 SamplesPerType[] = {1, 5, 3, 2, 4, 6, 7};
 		if(!isSynth && maskedType < std::size(SamplesPerType))
@@ -832,10 +850,10 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 			// TODO: Move octaves so that they align better (C-4 = lowest, we don't have access to the highest four octaves)
 			for(int octave = 4; octave < 10; octave++)
 			{
-				for(int note = 0, i = 12 * octave; note < 12; note++, i++)
+				for(int note = 0; note < 12; note++)
 				{
-					instr.Keyboard[i] = smp + OctSampleMap[numSamples - 2][octave - 4];
-					instr.NoteMap[i] = static_cast<uint8>(instr.NoteMap[i] + OctTransposeMap[numSamples - 2][octave - 4]);
+					instr.Keyboard[12 * octave + note] = smp + OctSampleMap[numSamples - 2][octave - 4];
+					instr.NoteMap[12 * octave + note] += OctTransposeMap[numSamples - 2][octave - 4];
 				}
 			}
 		} else if(maskedType == MMDInstrHeader::EXTSAMPLE)
@@ -844,21 +862,25 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 			instr.Transpose(-24);
 		} else if(!isSynth && hardwareMixSamples)
 		{
-			for(auto &note : instr.NoteMap)
+			for(int octave = 7; octave < 10; octave++)
 			{
-				int realNote = note + sampleHeader.sampleTranspose;
-				if(realNote >= NOTE_MIDDLEC + 24)
-					note -= static_cast<uint8>(((realNote - NOTE_MIDDLEC - 12) / 12) * 12);
+				for(int note = 0; note < 12; note++)
+				{
+					instr.NoteMap[12 * octave + note] -= static_cast<uint8>((octave - 6) * 12);
+				}
 			}
 		}
 
+		MMD0Sample sampleHeader;
+		sampleHeaderChunk.ReadStruct(sampleHeader);
+
 		// midiChannel = 0xFF == midi instrument but with invalid channel, midiChannel = 0x00 == sample-based instrument?
 		if(sampleHeader.midiChannel > 0 && sampleHeader.midiChannel <= 16)
 		{
 			instr.nMidiChannel = sampleHeader.midiChannel - 1 + MidiFirstChannel;
 			needInstruments = true;
 
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 			if(!isSynth)
 			{
 				auto &mixPlug = m_MixPlugins[numPlugins];
@@ -874,7 +896,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 
 				numPlugins++;
 			}
-#endif  // NO_VST
+#endif  // MPT_WITH_VST
 		}
 		if(sampleHeader.midiPreset > 0 && sampleHeader.midiPreset <= 128)
 		{
@@ -915,7 +937,6 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 		{
 			sampleIO |= SampleIO::stereoSplit;
 			length /= 2;
-			m_SongFlags.reset(SONG_ISAMIGA);  // Amiga resampler does not handle stereo samples
 		}
 		if(instrHeader.type & MMDInstrHeader::DELTA)
 		{
@@ -979,7 +1000,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 				ins.nGlobalVol = (instrExt.volume + 1u) / 2u;
 			if(size > offsetof(MMDInstrExt, midiBank))
 				ins.wMidiBank = instrExt.midiBank;
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 			if(ins.nMixPlug > 0)
 			{
 				PLUGINDEX plug = ins.nMixPlug - 1;
@@ -1008,7 +1029,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 					}
 				}
 			}
-#endif  // NO_VST
+#endif  // MPT_WITH_VST
 
 			ModSample &sample = Samples[ins.Keyboard[NOTE_MIDDLEC]];
 			sample.nFineTune = MOD2XMFineTune(instrExt.finetune);
@@ -1045,7 +1066,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 
 	// Setup a program change macro for command 1C (even if MIDI plugin is disabled, as otherwise these commands may act as filter commands)
 	m_MidiCfg.ClearZxxMacros();
-	strcpy(m_MidiCfg.szMidiSFXExt[0], "Cc z");
+	m_MidiCfg.SFx[0] = "Cc z";
 
 	file.Rewind();
 	PATTERNINDEX basePattern = 0;
@@ -1066,7 +1087,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 		uint32 preamp = 32;
 		if(version < 2)
 		{
-			if(songHeader.songLength > 256 || m_nChannels > 16)
+			if(songHeader.songLength > 256)
 				return false;
 			ReadOrderFromArray(order, songHeader.GetMMD0Song().sequence, songHeader.songLength);
 			for(auto &ord : order)
@@ -1075,7 +1096,9 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 			}
 
 			SetupMODPanning(true);
-			for(CHANNELINDEX chn = 0; chn < m_nChannels; chn++)
+			// With MED SoundStudio 1.03 it's possible to create MMD1 files with more than 16 channels.
+			const CHANNELINDEX numChannelVols = std::min(m_nChannels, CHANNELINDEX(16));
+			for(CHANNELINDEX chn = 0; chn < numChannelVols; chn++)
 			{
 				ChnSettings[chn].nVolume = std::min<uint8>(songHeader.trackVol[chn], 64);
 			}
@@ -1130,12 +1153,12 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 				if(!order.empty())
 					order.push_back(order.GetIgnoreIndex());
 
-				const size_t orderStart = order.size();
 				size_t readOrders = playSeq.length;
 				if(!file.CanRead(readOrders))
 					LimitMax(readOrders, file.BytesLeft());
 				LimitMax(readOrders, ORDERINDEX_MAX);
 
+				size_t orderStart = order.size();
 				order.reserve(orderStart + readOrders);
 				for(size_t ord = 0; ord < readOrders; ord++)
 				{
@@ -1146,7 +1169,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 					}
 				}
 				if(playSeq.name[0])
-					order.SetName(mpt::ToUnicode(mpt::Charset::ISO8859_1, mpt::String::ReadAutoBuf(playSeq.name)));
+					order.SetName(mpt::ToUnicode(mpt::Charset::Amiga_no_C1, mpt::String::ReadAutoBuf(playSeq.name)));
 
 				// Play commands (jump / stop)
 				if(playSeq.commandTableOffset > 0 && file.Seek(playSeq.commandTableOffset))
@@ -1189,25 +1212,23 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 
 		// For MED, this affects both volume and pitch slides
 		m_SongFlags.set(SONG_FASTVOLSLIDES, !(songHeader.flags & MMDSong::FLAG_STSLIDE));
-		m_playBehaviour.set(kST3PortaSampleChange);
-		m_playBehaviour.set(kFT2PortaNoNote);
 
 		if(expData.songNameOffset && file.Seek(expData.songNameOffset))
 		{
 			file.ReadString<mpt::String::maybeNullTerminated>(m_songName, expData.songNameLength);
 			if(numSongs > 1)
-				order.SetName(mpt::ToUnicode(mpt::Charset::ISO8859_1, m_songName));
+				order.SetName(mpt::ToUnicode(mpt::Charset::Amiga_no_C1, m_songName));
 		}
 		if(expData.annoLength > 1 && file.Seek(expData.annoText))
 		{
 			m_songMessage.Read(file, expData.annoLength - 1, SongMessage::leAutodetect);
 		}
 
-#ifndef NO_VST
+#ifdef MPT_WITH_VST
 		// Read MIDI messages
 		if(expData.midiDumpOffset && file.Seek(expData.midiDumpOffset) && file.CanRead(8))
 		{
-			uint16 numDumps = std::min(file.ReadUint16BE(), static_cast<uint16>(std::size(m_MidiCfg.szMidiZXXExt)));
+			uint16 numDumps = std::min(file.ReadUint16BE(), static_cast<uint16>(m_MidiCfg.Zxx.size()));
 			file.Skip(6);
 			if(file.CanRead(numDumps * 4))
 			{
@@ -1221,18 +1242,19 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 					file.ReadStruct(dumpHeader);
 					if(!file.Seek(dumpHeader.dataPointer) || !file.CanRead(dumpHeader.length))
 						continue;
-					auto &macro = m_MidiCfg.szMidiZXXExt[dump];
-					auto length = std::min(static_cast<size_t>(dumpHeader.length), std::size(macro) / 2u);
+					std::array<char, kMacroLength> macro{};
+					auto length = std::min(static_cast<size_t>(dumpHeader.length), macro.size() / 2u);
 					for(size_t i = 0; i < length; i++)
 					{
 						const uint8 byte = file.ReadUint8(), high = byte >> 4, low = byte & 0x0F;
 						macro[i * 2] = high + (high < 0x0A ? '0' : 'A' - 0x0A);
 						macro[i * 2 + 1] = low + (low < 0x0A ? '0' : 'A' - 0x0A);
 					}
+					m_MidiCfg.Zxx[dump] = std::string_view{macro.data(), length * 2};
 				}
 			}
 		}
-#endif
+#endif  // MPT_WITH_VST
 
 		if(expData.mmdInfoOffset && file.Seek(expData.mmdInfoOffset) && file.CanRead(12))
 		{
@@ -1280,7 +1302,6 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 		}
 
 		PATTERNINDEX numPatterns = songHeader.numBlocks;
-		LimitMax(numPatterns, static_cast<PATTERNINDEX>(PATTERNINDEX_INVALID - basePattern));
 		Patterns.ResizeArray(basePattern + numPatterns);
 		for(PATTERNINDEX pat = 0; pat < numPatterns; pat++)
 		{
@@ -1317,9 +1338,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 					file.Seek(patHeader.blockInfoOffset);
 					MMDBlockInfo blockInfo;
 					file.ReadStruct(blockInfo);
-					if(blockInfo.nameLength
-					   && blockInfo.nameOffset
-					   && file.Seek(blockInfo.nameOffset))
+					if(file.Seek(blockInfo.nameOffset))
 					{
 						// We have now chased four pointers to get this far... lovely format.
 						file.ReadString<mpt::String::maybeNullTerminated>(patName, blockInfo.nameLength);
@@ -1347,7 +1366,9 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 				ModCommand *m = pattern.GetpModCommand(row, 0);
 				for(CHANNELINDEX chn = 0; chn < numTracks; chn++, m++)
 				{
+					const auto oldCmd = std::make_pair(m->command, m->param);
 					int note = NOTE_NONE;
+					uint8 cmd = 0;
 					if(version < 1)
 					{
 						const auto [noteInstr, instrCmd, param] = file.ReadArray<uint8, 3>();
@@ -1357,7 +1378,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 
 						m->instr = (instrCmd >> 4) | ((noteInstr & 0x80) >> 3) | ((noteInstr & 0x40) >> 1);
 
-						m->command = instrCmd & 0x0F;
+						cmd = instrCmd & 0x0F;
 						m->param = param;
 					} else
 					{
@@ -1370,7 +1391,7 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 							m->note = NOTE_NOTECUT;
 
 						m->instr = instr & 0x3F;
-						m->command = command;
+						cmd = command;
 						m->param = param1;
 					}
 					// Octave wrapping for 4-channel modules (TODO: this should not be set because of synth instruments)
@@ -1379,7 +1400,23 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 
 					if(note >= NOTE_MIN && note <= NOTE_MAX)
 						m->note = static_cast<ModCommand::NOTE>(note);
-					ConvertMEDEffect(*m, is8Ch, bpmMode, rowsPerBeat, volHex);
+					const auto extraCmd = ConvertMEDEffect(*m, cmd, is8Ch, bpmMode, rowsPerBeat, volHex);
+
+					if(oldCmd.first == CMD_XPARAM)
+					{
+						// Restore X-Param if it was overwritten by an empty effect, or restrict to 8-bit value if this cell was overwritten with a "useful" effect
+						if(m->command == CMD_NONE)
+							m->SetEffectCommand(oldCmd);
+						else if(row > 0)
+							pattern.GetpModCommand(row - 1, chn)->param = Util::MaxValueOfType(m->param);
+					}
+					if(extraCmd.first != CMD_NONE)
+					{
+						if(row < (numRows - 1))
+							pattern.GetpModCommand(row + 1, chn)->SetEffectCommand(extraCmd);
+						else
+							m->param = Util::MaxValueOfType(m->param);  // No space :(
+					}
 				}
 			}
 		}
@@ -1434,10 +1471,10 @@ bool CSoundFile::ReadMED(FileReader &file, ModLoadingFlags loadFlags)
 	case 3: madeWithTracker = MPT_ULITERAL("OctaMED Soundstudio (MMD3)"); break;
 	}
 
-	m_modFormat.formatName = mpt::format(MPT_USTRING("OctaMED (MMD%1)"))(version);
+	m_modFormat.formatName = MPT_UFORMAT("OctaMED (MMD{})")(version);
 	m_modFormat.type = MPT_USTRING("med");
 	m_modFormat.madeWithTracker = madeWithTracker;
-	m_modFormat.charset = mpt::Charset::ISO8859_1;
+	m_modFormat.charset = mpt::Charset::Amiga_no_C1;
 
 	return true;
 }
