@@ -66,7 +66,7 @@ struct SFXSampleHeader
 		mptSmp.Initialize(MOD_TYPE_MOD);
 		mptSmp.nLength = length;
 		mptSmp.nFineTune = MOD2XMFineTune(finetune);
-		mptSmp.nVolume = 4u * std::min<uint8>(volume, 64);
+		mptSmp.nVolume = 4u * std::min(volume.get(), uint8(64));
 
 		SmpLength lStart = loopStart;
 		SmpLength lLength = loopLength * 2u;
@@ -105,17 +105,17 @@ static uint8 ClampSlideParam(uint8 value, uint8 lowNote, uint8 highNote)
 	uint16 lowPeriod, highPeriod;
 
 	if(lowNote  < highNote &&
-	   lowNote  >= 36 + NOTE_MIN &&
-	   highNote >= 36 + NOTE_MIN &&
-	   lowNote  < CountOf(ProTrackerPeriodTable) + 36 + NOTE_MIN &&
-	   highNote < CountOf(ProTrackerPeriodTable) + 36 + NOTE_MIN)
+	   lowNote  >= 24 + NOTE_MIN &&
+	   highNote >= 24 + NOTE_MIN &&
+	   lowNote  < std::size(ProTrackerPeriodTable) + 24 + NOTE_MIN &&
+	   highNote < std::size(ProTrackerPeriodTable) + 24 + NOTE_MIN)
 	{
-		lowPeriod  = ProTrackerPeriodTable[lowNote - 36 - NOTE_MIN];
-		highPeriod = ProTrackerPeriodTable[highNote - 36 - NOTE_MIN];
+		lowPeriod  = ProTrackerPeriodTable[lowNote - 24 - NOTE_MIN];
+		highPeriod = ProTrackerPeriodTable[highNote - 24 - NOTE_MIN];
 
 		// with a fixed speed of 6 ticks/row, and excluding the first row,
 		// 1xx/2xx param has a max value of (low-high)/5 to avoid sliding too far
-		return std::min<uint8>(value, static_cast<uint8>((lowPeriod - highPeriod) / 5));
+		return std::min(value, static_cast<uint8>((lowPeriod - highPeriod) / 5));
 	}
 
 	return 0;
@@ -200,7 +200,6 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 
 	// Setup channel pan positions and volume
 	SetupMODPanning(true);
-	m_playBehaviour.set(kMODIgnorePanning);
 
 	uint32 invalidChars = 0;
 	for(SAMPLEINDEX smp = 1; smp <= m_nSamples; smp++)
@@ -221,13 +220,13 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 		}
 		if(invalidChars >= 128)
 			return false;
-		mpt::String::Read<mpt::String::spacePadded>(m_szNames[smp], sampleHeader.name);
+		m_szNames[smp] = mpt::String::ReadBuf(mpt::String::spacePadded, sampleHeader.name);
 	}
 
 	// Broken conversions of the "Operation Stealth" soundtrack (BOND23 / BOND32)
 	// There is a converter that shifts all note values except FFFD (empty note) to the left by 1 bit,
 	// but it should not do that for FFFE (STP) notes - as a consequence, they turn into pattern breaks (FFFC).
-	const bool fixPatternBreaks = !strcmp(m_szNames[1], "BASSE2.AMI") || !strcmp(m_szNames[1], "PRA1.AMI");
+	const bool fixPatternBreaks = (m_szNames[1] == "BASSE2.AMI") || (m_szNames[1] == "PRA1.AMI");
 
 	SFXOrderHeader orderHeader;
 	if(!file.ReadStruct(orderHeader) || !orderHeader.IsValid())
@@ -255,6 +254,7 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 	uint8 lastNote[4] = {0};
 	uint8 slideTo[4] = {0};
 	uint8 slideRate[4] = {0};
+	uint8 version = 0;
 
 	// Reading patterns
 	if(loadFlags & loadPatternData)
@@ -269,12 +269,11 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 
 		for(ROWINDEX row = 0; row < 64; row++)
 		{
-			PatternRow rowBase = Patterns[pat].GetpModCommand(row, 0);
+			auto rowBase = Patterns[pat].GetRow(row);
 			for(CHANNELINDEX chn = 0; chn < 4; chn++)
 			{
 				ModCommand &m = rowBase[chn];
-				uint8 data[4];
-				file.ReadArray(data);
+				auto data = file.ReadArray<uint8, 4>();
 
 				if(data[0] == 0xFF)
 				{
@@ -292,20 +291,26 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 						continue;
 					case 0xFC: // BRK (pattern break)
 						m.command = CMD_PATTERNBREAK;
+						version = 9;
 						continue;
 					}
 				}
 
-				ReadMODPatternEntry(data, m);
+				const auto [command, param] = ReadMODPatternEntry(data, m);
 				if(m.note != NOTE_NONE)
 				{
 					lastNote[chn] = m.note;
 					slideRate[chn] = 0;
+					if(m.note < NOTE_MIDDLEC - 12)
+					{
+						version = std::max(version, uint8(8));
+					}
 				}
 
-				if(m.command || m.param)
+				if(command || param)
 				{
-					switch(m.command)
+					m.param = param;
+					switch(command)
 					{
 					case 0x1: // Arpeggio
 						m.command = CMD_ARPEGGIO;
@@ -322,7 +327,7 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 							m.param &= 0x0F;
 						} else
 						{
-							m.command = m.param = 0;
+							m.command = CMD_NONE;
 						}
 						break;
 
@@ -330,41 +335,37 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 						// Give precedence to 7xy/8xy slides
 						if(slideRate[chn])
 						{
-							m.command = m.param = 0;
+							m.command = CMD_NONE;
 							break;
 						}
-						m.command = CMD_MODCMDEX;
-						m.param = 0;
+						m.SetEffectCommand(CMD_MODCMDEX, 0x00);
 						break;
 
 					case 0x4: // Disable LED filter
 						// Give precedence to 7xy/8xy slides
 						if(slideRate[chn])
 						{
-							m.command = m.param = 0;
+							m.command = CMD_NONE;
 							break;
 						}
-						m.command = CMD_MODCMDEX;
-						m.param = 1;
+						m.SetEffectCommand(CMD_MODCMDEX, 0x01);
 						break;
 
 					case 0x5: // Increase volume
 						if(m.instr)
 						{
-							m.command = CMD_VOLUME;
-							m.param = std::min(ModCommand::PARAM(0x3F), static_cast<ModCommand::PARAM>((Samples[m.instr].nVolume / 4u) + m.param));
+							m.SetEffectCommand(CMD_VOLUME, std::min(ModCommand::PARAM(0x3F), static_cast<ModCommand::PARAM>((Samples[m.instr].nVolume / 4u) + m.param)));
 
 							// Give precedence to 7xy/8xy slides (and move this to the volume column)
 							if(slideRate[chn])
 							{
-								m.volcmd = VOLCMD_VOLUME;
-								m.vol = m.param;
-								m.command = m.param = 0;
+								m.SetVolumeCommand(VOLCMD_VOLUME, m.param);
+								m.command = CMD_NONE;
 								break;
 							}
 						} else
 						{
-							m.command = m.param = 0;
+							m.command = CMD_NONE;
 						}
 						break;
 
@@ -380,33 +381,33 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 							// Give precedence to 7xy/8xy slides (and move this to the volume column)
 							if(slideRate[chn])
 							{
-								m.volcmd = VOLCMD_VOLUME;
-								m.vol = m.param;
-								m.command = m.param = 0;
+								m.SetVolumeCommand(VOLCMD_VOLUME, m.param);
+								m.command = CMD_NONE;
 								break;
 							}
 						} else
 						{
-							m.command = m.param = 0;
+							m.command = CMD_NONE;
 						}
 						break;
 
 					case 0x7: // 7xy: Slide down x semitones at speed y
 						slideTo[chn] = lastNote[chn] - (m.param >> 4);
 
-						m.command = CMD_PORTAMENTODOWN;
 						slideRate[chn] = m.param & 0xF;
-						m.param = ClampSlideParam(slideRate[chn], slideTo[chn], lastNote[chn]);
+						m.SetEffectCommand(CMD_PORTAMENTODOWN, ClampSlideParam(slideRate[chn], slideTo[chn], lastNote[chn]));
 						break;
 
 					case 0x8: // 8xy: Slide up x semitones at speed y
 						slideTo[chn] = lastNote[chn] + (m.param >> 4);
 
-						m.command = CMD_PORTAMENTOUP;
 						slideRate[chn] = m.param & 0xF;
-						m.param = ClampSlideParam(slideRate[chn], lastNote[chn], slideTo[chn]);
+						m.SetEffectCommand(CMD_PORTAMENTOUP, ClampSlideParam(slideRate[chn], lastNote[chn], slideTo[chn]));
 						break;
 
+					case 0x9: // 9xy: Auto slide
+						version = std::max(version, uint8(8));
+						[[fallthrough]];
 					default:
 						m.command = CMD_NONE;
 						break;
@@ -442,6 +443,10 @@ bool CSoundFile::ReadSFX(FileReader &file, ModLoadingFlags loadFlags)
 		}
 	}
 
+	m_modFormat.formatName = m_nSamples == 15 ? MPT_UFORMAT("SoundFX 1.{}")(version) : U_("SoundFX 2.0 / MultiMedia Sound");
+	m_modFormat.type = m_nSamples == 15 ? UL_("sfx") : UL_("sfx2");
+	m_modFormat.charset = mpt::Charset::Amiga_no_C1;
+
 	return true;
 }
 
